/*
    Fichier cyladdval.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Programme permettant de mettre  jour la variable d'un fichier
    cyloop lorsqu'elle mmorise les valeurs de cette variable.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "limites.h"
#include "messages.h"
#include "majcycle.h"


// prototypes
void pondere (int coef_pond);
void addval (char *chaineval, uint16 num_instant, int majcmoy);
void addvalint32 (int32 val, uint16 num_instant, int majcmoy);
void addvalfloat (float val, uint16 num_instant, int majcmoy);
int  testval (char *valeur);



// variables globales (pour viter des tas de passages de paramtre)
FILE        *descfic;  // descripteur du fichier cyloop
desc_cyloop entetefic; // entte du fichier cyloop
desc_var    entetevar; // entte de la variable du fichier cyloop



// programme principal

main (int narg, char *varg[])
{
    char      *nomfic;       // nom du fichier cyloop
    struct tm stm;           // structure contenant la date et l'heure courante
    uint32    timep;         // date et heure dans un uint32
    int       maj_entetefic; // indicateur pour mise  jour fichier
    uint16    num_instant;     // numro de la donne  incrmenter
    uint32    duree_cycle;   // dure du cycle


    // mmorisation du nom de la commande
    memcom (*varg);

    // test des paramtres
    if (--narg > 1 && testval (varg [2]))
        // rcupration du nom de fichier et rajout ventuel du suffixe
        nomfic = recup_nomfic (varg [1]);
    else
    {
        // "Nom de fichier manquant ou valeur de la variable manquante"
        // "Syntaxe : %s nom_fichier_cyloop valeur"
        affiche_err ("MANQUE_NOM_OU_VAR");
        psyntaxe ("SYNT_CYLADDVAL");
    }

    // ouverture en lecture / criture du fichier cyloop
    descfic = fopen (nomfic, "r+");

    if (! descfic)
        // "Fichier_cyloop %s inexistant ou protg en lecture ou criture"
        err_fatale_arg ("ACCES_CYLOOP_RW",
                                                                   nomfic);

    // lecture de l'entte du fichier cyloop
    fread (&entetefic, 16, 1, descfic);

    // vrification de la signature
    if (entetefic.signature [0] != 'c' || entetefic.signature [1] != 'y'
                                       || entetefic.signature [2] != 'l')
    {
        // "Le fichier %s n'est pas un fichier_cyloop"
        aff_err_arg ("FICH_NON_CYLOOP", nomfic);
        fclose (descfic);
        return;
    }

    // vrification de la valeur nb_var
    if (entetefic.nb_var)
    {
        // "Ce fichier cyloop utilise une liste de variables"
        // "utiliser la version complte de cyloop"
        affiche_err ("CYLOOP_LISTEVAR");
        affiche_err ("UTIL_VCOMPLETE");
        fclose (descfic);
        return;
    }

    // lecture de la partie fixe de l'entte de la variable du fichier cyloop
    fread (&entetevar, 6, 1, descfic);

    // vrification de la nature de la variable
    if (entetevar.typedon & compteur)
    {
        // "La variable du fichier cyloop est un compteur : utiliser cylincr"
        affiche_err ("VAR_COMPTEUR");
        fclose (descfic);
        return;
    }

    // fin de lecture de l'entte de la variable si ncessaire
    if (! (entetevar.typedon & cmoy_sep))
        fread (&entetevar.coef_moy, 2, 1, descfic);

    // rcuprer la date et l'heure courante
    time (&timep);

    // si le dbut du cycle est ultrieur  l'instant prsent
    if (timep < entetefic.deb_cycle)
    {
        fclose (descfic);

        // "Rgression dans le temps : une intervention manuelle "
        // "sur le fichier cyloop ou l'horloge est ncessaire"
        affiche_err ("REGRES_TEMPS");
        err_fatale ("REGRES_TEMPS2");
    }

    // initialisation indicateurs de mise  jour du fichier
    maj_entetefic = NON;

    // si on utilise l'heure locale
    if (! (entetefic.caract_gen & temps_utc))
    {
        // rcuprer le dtail de la date et de l'heure locale
        localtime_r (&timep, &stm);

        // si changement d'heure t/hiver
        if ((entetefic.etat_cyloop & heure_ete) != stm.tm_isdst)
            // prendre en compte ce changement
            maj_entetefic = maj_type_heure (stm.tm_isdst, timep);
    }

    // si changement de cycle
    if (entetefic.deb_cycle + entetefic.duree_cycle <= timep)
    {
        // prise en compte de ce changement de cycle
        changement_cycle (timep);

        // une manire de mmoriser un dbut de cycle
        entetevar.der_enreg = 0;

        // on reportera la mise  jour dans le fichier cyloop
        maj_entetefic = OUI;
    }

    // si mise  jour de l'entte du fichier cyloop ncessaire
    if (maj_entetefic)
    {
        // recopier la nouvelle entte
        rewind (descfic);
        fwrite (&entetefic, 16, 1, descfic);
    }

    // rcupration de la dure du cycle actuel
    duree_cycle = duree_cycle_actuel (&timep);

    // calcul de l'instant du cycle  mettre  jour
    num_instant = calcul_instant (timep, duree_cycle);

    // si on a chang de cycle et une seule valeur coef_moy pour la variable
    if (! entetevar.der_enreg && !(entetevar.typedon & cmoy_sep))
    {
        // mettre  jour cette valeur
        if (entetevar.ponderation)
            entetevar.coef_moy += 256;
        else
            entetevar.coef_moy ++;

        // y compris dans le fichier cyloop
        fseek (descfic, 22, SEEK_SET);
        fwrite (&entetevar.coef_moy, 2, 1, descfic);
    }

    // si c'est la premire mise  jour effectue durant ce cycle
    // ou si on passe  un autre instant du mme cycle
    if (entetevar.der_enreg < num_instant)
    {
        // mettre  jour le numro de l'instant du cycle trait
        entetevar.der_enreg = num_instant;

        // y compris dans le fichier cyloop
        fseek (descfic, 16, SEEK_SET);
        fwrite (&entetevar.der_enreg, 2, 1, descfic);

        // mettre  jour la valeur de la variable et si ncessaire
        // le champ coef_moy pour l'instant du cycle concern
        addval (varg [2], num_instant, OUI);
    }
    // sinon, si on a dj mis  jour la variable pour cet instant du cycle
    else if (entetevar.der_enreg == num_instant)
    {
        // si on est autoris  le mettre  jour plusieur
        // fois pour chacun des instants du cycle
        if (entetevar.typedon & cumul_donnees)
            // mettre  jour la valeur de la variable et si ncessaire
            // le champ coef_moy pour l'instant du cycle concern
            addval (varg [2], num_instant, entetevar.typedon & cpt_toute_don);
    }
    // sinon, il y a rgression dans le temps
    else
    {
        // vrifier si elle est suprieure  2 heures
        if (calcul_instant (timep + 7200, duree_cycle) < entetevar.der_enreg)
        {
            // vrifier si elle est suprieure  1 journe
            // et afficher le message d'erreur appropri
            if (calcul_instant (timep + secjour, duree_cycle) <
                                                 entetevar.der_enreg)
            {
                // "Forte rgression dans le temps : une intervention manuelle "
                // "sur le fichier cyloop ou l'horloge est ncessaire"
                affiche_err ("FORTE_REGR_TEMPS");
                affiche_err ("REGRES_TEMPS2");
            }
            else
              // "Rgression dans le temps : vrifier l'horloge de l'ordinateur"
                affiche_err ("REGR_VERIF_HORL");
        }
    }

    // le fichier cyloop a t mis  jour
    fclose (descfic);
}



// test de la valeur passe en paramtre

int testval (char *valeur)
{
    // saut du signe - ventuel
    if (*valeur == '-')
        valeur++;

    // si la valeur commence par des chiffres
    if ('0' <= *valeur && *valeur <= '9')
    {
        // les parcourrir
        do
            valeur++;
        while ('0' <= *valeur && *valeur <= '9');

        // si plus rien aprs les chiffres
        if (! *valeur)
            // la valeur est un entier
            return OUI;

        // si partie dcimale
        if (*valeur == '.' || *valeur == ',')
        {
            // la parcourir
            do
                valeur++;
            while ('0' <= *valeur && *valeur <= '9');
        }

        // si exposant
        if (tolower (*valeur) == 'e')
        {
            // on va l'analyser
            valeur ++;

            // saut du signe - ventuel
            if (*valeur == '-')
                valeur++;

            // si l'exposant contient des chiffres
            if ('0' <= *valeur && *valeur <= '9')
            {
                // les parcourrir
                do
                    valeur++;
                while ('0' <= *valeur && *valeur <= '9');
            }
        }

        // succs de l'analyse si l'exploration de la valeur est termin
        return (*valeur == '\0');
    }
    // sinon, contenu de la valeur incorrect
    else
        return NON;
}



// calcul de pondration sur les cycles antrieurs

void pondere (int coef_pond)
{
    uint16 coef_moy; // coef_moy si distinct pour chaque instant du cycle
    int32  valeur32b;  // valeur  pondrer si elle est entire sur 32 bits
    float  valeurf;    // valeur  pondrer si elle est en virgule flottante
    uint32 valmaj;     // valeur sur 32 bits pour calcul de pondration
    uint32 positfic;   // positionnement sur une nouvelle valeur du cycle
    int    tailledon;  // nombre d'octets des donnes pour un instant du cycle
    int    i;          // compteur


    // nombre d'octets pour chaque instant du cycle
    if (entetevar.typedon & mem_min_max)
        tailledon = 12;
    else
        tailledon = 4;

    // si un coefficient coef_moy par instant du cycle
    if (entetevar.typedon & cmoy_sep)
    {
        // corriger le calcul de tailledon
        tailledon = tailledon + 2;

        // position des premires donnes dans le fichier
        positfic = 22;

        // si valeurs en virgule flottante
        if (entetevar.typedon & val_float)
        {
            // mise  jour des valeurs de coef_moy et de la variable
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // positionnement dans le fichier
                fseek (descfic, positfic, SEEK_SET);

                // rcupration de la valeur du coef_moy
                fread (&coef_moy, 2, 1, descfic);

                // rcupration de la valeur de la variable en virgule flottante
                fread (&valeurf, 4, 1, descfic);

                // calcul de pondration pour coef_moy
                // le calcul de multiplication et division est fait sur 32 bits
                valmaj = coef_moy;
                coef_moy = (valmaj * coef_pond) >> 8;

                // calcul de la nouvelle valeur
                valeurf = valeurf * (coef_pond / 256.0);

                // mettre  jour les valeurs dans le fichier
                fseek (descfic, -6, SEEK_CUR);
                fwrite (&coef_moy, 2, 1, descfic);
                fwrite (&valeurf, 4, 1, descfic);

                // position des donnes pour l'instant suivant dans le fichier
                positfic = positfic + tailledon;
            }
        }
        // sinon (valeurs de la variable entire sur 32 bits)
        else
        {
            // mise  jour des valeurs de coef_moy et de la variable 32 bits
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // positionnement dans le fichier
                fseek (descfic, positfic, SEEK_SET);

                // rcupration de la valeur du coef_moy
                fread (&coef_moy, 2, 1, descfic);

                // rcupration de la valeur de la variable sur 32 bits
                fread (&valeur32b, 4, 1, descfic);

                // calcul de pondration pour coef_moy
                // le calcul de multiplication et division est fait sur 32 bits
                valmaj = coef_moy;
                coef_moy = (valmaj * coef_pond) >> 8;

                // calcul de pondration pour la valeur de la variable
                // si risque de dbordement lors du calcul
                if (valeur32b & 0xFF000000)
                    // faire la division d'abord
                    valeur32b = (valeur32b >> 8) * coef_pond;

                // sinon faire la multiplication d'abord
                else
                    valeur32b = (valeur32b * coef_pond) >> 8;

                // mettre  jour les valeurs dans le fichier
                fseek (descfic, -6, SEEK_CUR);
                fwrite (&coef_moy, 2, 1, descfic);
                fwrite (&valeur32b, 4, 1, descfic);

                // position des donnes pour l'instant suivant dans le fichier
                positfic = positfic + tailledon;
            }
        }
    }
    // sinon (coefficient coef_moy global  la variable)
    else
    {
        // position des premires donnes dans le fichier
        positfic = 24;

        // mettre  jour ce coefficient
        // le calcul de multiplication et division est fait sur 32 bits
        valmaj = entetevar.coef_moy;
        entetevar.coef_moy = (valmaj * coef_pond) >> 8;

        // ce coefficient sera mis  jour plus tard dans le fichier

        // si valeurs en virgule flottante
        if (entetevar.typedon & val_float)
        {
            // mise  jour des valeurs en virgule flottante
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // positionnement dans le fichier
                fseek (descfic, positfic, SEEK_SET);

                // rcupration de la valeur de la variable en virgule flottante
                fread (&valeurf, 4, 1, descfic);

                // calcul de la nouvelle valeur
                valeurf = valeurf * (coef_pond / 256.0);

                // mettre  jour la valeur de la variable dans le fichier
                fseek (descfic, -4, SEEK_CUR);
                fwrite (&valeurf, 4, 1, descfic);

                // position des donnes pour l'instant suivant dans le fichier
                positfic = positfic + tailledon;
            }
        }
        // sinon (valeurs de la variable entire sur 32 bits)
        else
        {
            // mise  jour des valeurs entires sur 32 bits
            for (i = entetefic.nb_don_cycle; i > 0; i--)
            {
                // positionnement dans le fichier
                fseek (descfic, positfic, SEEK_SET);

                // rcupration de la valeur de la variable sur 32 bits
                fread (&valeur32b, 4, 1, descfic);

                // si risque de dbordement lors du calcul
                if (valeur32b & 0xFF000000)
                    // faire la division d'abord
                    valeur32b = (valeur32b >> 8) * coef_pond;

                // sinon faire la multiplication d'abord
                else
                    valeur32b = (valeur32b * coef_pond) >> 8;

                // mettre  jour la valeur de la variable dans le fichier
                fseek (descfic, -4, SEEK_CUR);
                fwrite (&valeur32b, 4, 1, descfic);

                // position des donnes pour l'instant suivant dans le fichier
                positfic = positfic + tailledon;
            }
        }
    }
}



// rajoute la valeur contenue dans chaineval  l'instant num_instant du cycle

void addval (char *chaineval, uint16 num_instant, int majcmoy)
{
    // selon le type des donnes mmorises
    if (entetevar.typedon & val_float)
        // appeler la bonne fonction d'ajout
        addvalfloat (atof (chaineval), num_instant, majcmoy);
    else
        addvalint32 (atol (chaineval), num_instant, majcmoy);
}



// rajoute la valeur val (entire 32 bits)  l'instant num_instant du cycle

void addvalint32 (int32 val, uint16 num_instant, int majcmoy)
{
    long   posval;     // position de la valeur  mettre  jour dans le fichier
    uint16 coef_moy; // coefficient coef_moy pour la valeur  mettre  jour
    int32  valeur;     // valeur  mettre  jour
    int32  nouv_val;   // valeur mise  jour
    int32  ajout;      // val  rajouter  la variable avec coef multiplicateur
    int32  min, max;   // valeurs min et max mmorises
    int    majfich;    // indique si on devra mettre  jour le fichier
    int    maj_minmax; // indique si on devra mettre  jour la valeur min ou max


    // initialisation
    majfich = OUI;
    maj_minmax = NON;

    // pour le calcul de position, le premier instant
    // du cycle n'entrainera pas de dcalage
    num_instant --;

    // calcul de la position dans le fichier de la valeur  mettre  jour
    posval = 24;

    if (entetevar.typedon & cmoy_sep)
        posval += num_instant * 2;

    if (entetevar.typedon & mem_min_max)
        posval += num_instant * 12;
    else
        posval += num_instant * 4;

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // rcuprer et mettre  jour le coefficient coef_moy
        fread (&coef_moy, 2, 1, descfic);

        if (entetevar.ponderation)
            coef_moy += 256;
        else
            coef_moy += 1;
    }
    // sinon, se positionner sur la valeur de la variable
    else
        fseek (descfic, posval, SEEK_SET);

    // rcupration de la valeur de la variable
    fread (&valeur, 4, 1, descfic);

    // si valeurs min et max  mettre  jour
    if (entetevar.typedon & mem_min_max)
    {
        // rcuprer les valeurs mmorises dans le fichier
        fread (&min, 4, 1, descfic);
        fread (&max, 4, 1, descfic);

        // les actualiser si ncessaire
        if (val < min)
        {
            min = val;
            maj_minmax = OUI;
        }

        // pas de else if, la premire fois, il faut actualiser min et max
        if (val > max)
        {
            max = val;
            maj_minmax = OUI;
        }
    }

    // si calcul de pondration
    if (entetevar.ponderation)
    {
        // calculer la valeur  ajouter
        ajout = val * 256;

        // si dbordement numrique lors de ce calcul
        if (ajout / 256 != val)
        {
            // "Valeur  rajouter trop grande"
            // "Il faudrait utiliser une variable en virgule flottante"
            affiche_err ("VAL_TROP_GRANDE");
            affiche_err ("BESOIN_VARFLOAT");

            // on ajoutera une valeur plus petite
            if (val > 0)
                ajout = max_int32;
            else
                ajout = min_int32;
        }
    }
    // sinon, simple ajout de la valeur passe en paramtre
    else
        ajout = val;

    // calculer la nouvelle valeur
    nouv_val = valeur + ajout;

    // si variable et valeur  rajouter du mme signe
    if ((float) valeur * (float) ajout > 0)
    {
        // et que le signe a chang lors de l'addition
        if ((float) nouv_val * (float) ajout <= 0)
        {
            // nouvelle valeur de la variable
            if (valeur < 0)
                nouv_val = min_int32;
            else
                nouv_val = max_int32;

            // si on tait dj sur la valeur limite
            if (valeur == nouv_val)
            {
                // message d'erreur adapt
                // "Rappel : Dbordement numrique pour la variable"
                affiche_err ("PAR_DEBORD_NUM");

                // pas la peine de mettre  jour le fichier pour la variable
                majfich = NON;
            }
            // sinon
            else
            {
                // message d'erreur
                // "Dbordement numrique pour la variable"
                affiche_err ("DEBORD_NUMERIQUE");
            }

            // fin du message d'erreur
            // "Il faudrait la convertir en nombre flottant"
            affiche_err ("BESOIN_CONVFLOAT");
        }
    }

    // prise en compte de la nouvelle valeur
    valeur = nouv_val;

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // mettre  jour le coefficient coef_moy
        fwrite (&coef_moy, 2, 1, descfic);
    }
    // sinon
    else if (majfich | maj_minmax)
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval, SEEK_SET);
    }

    // mettre  jour si ncessaire la valeur de la variable
    if (majfich | maj_minmax)
        fwrite (&valeur, 4, 1, descfic);

    // mettre  jour si ncessaire la valeur min ou max
    if (maj_minmax)
    {
        fwrite (&min, 4, 1, descfic);
        fwrite (&max, 4, 1, descfic);
    }
}



// rajoute la valeur val (flottante)  l'instant num_instant du cycle

void addvalfloat (float val, uint16 num_instant, int majcmoy)
{
    long   posval;     // position de la valeur  mettre  jour dans le fichier
    uint16 coef_moy; // coefficient coef_moy pour la valeur  mettre  jour
    float  valeur;     // valeur  mettre  jour
    float  min, max;   // valeurs min et max mmorises
    int    maj_minmax; // indique si on devra mettre  jour la valeur min ou max


    // initialisation
    maj_minmax = NON;

    // pour le calcul de position, le premier instant
    // du cycle n'entrainera pas de dcalage
    num_instant --;

    // calcul de la position dans le fichier de la valeur  mettre  jour
    posval = 24;

    if (entetevar.typedon & cmoy_sep)
        posval += num_instant * 2;

    if (entetevar.typedon & mem_min_max)
        posval += num_instant * 12;
    else
        posval += num_instant * 4;

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // rcuprer et mettre  jour le coefficient coef_moy
        fread (&coef_moy, 2, 1, descfic);

        if (entetevar.ponderation)
            coef_moy += 256;
        else
            coef_moy += 1;
    }
    // sinon, se positionner sur la valeur de la variable
    else
        fseek (descfic, posval, SEEK_SET);

    // rcupration de la valeur de la variable
    fread (&valeur, 4, 1, descfic);

    // calculer sa nouvelle valeur (pas de test de dbordement ici)
    valeur += val;

    // si valeurs min et max  mettre  jour
    if (entetevar.typedon & mem_min_max)
    {
        // rcuprer les valeurs mmorises dans le fichier
        fread (&min, 4, 1, descfic);
        fread (&max, 4, 1, descfic);

        // les actualiser si ncessaire
        if (val < min)
        {
            min = val;
            maj_minmax = OUI;
        }

        // pas de else if, la premire fois, il faut actualiser min et max
        if (val > max)
        {
            max = val;
            maj_minmax = OUI;
        }
    }

    // si coefficient coef_moy  mettre  jour
    if (majcmoy && (entetevar.typedon & cmoy_sep))
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval - 2, SEEK_SET);

        // mettre  jour le coefficient coef_moy
        fwrite (&coef_moy, 2, 1, descfic);
    }
    // sinon
    else
    {
        // se positionner dans le fichier sur la zone  modifier
        fseek (descfic, posval, SEEK_SET);
    }

    // mettre  jour la valeur de la variable
    fwrite (&valeur, 4, 1, descfic);

    // mettre  jour si ncessaire la valeur min ou max
    if (maj_minmax)
    {
        fwrite (&min, 4, 1, descfic);
        fwrite (&max, 4, 1, descfic);
    }
}
